@ECHO OFF

:: released 2010 March 24

:: NAME
::
:: 	ROBOCOPY_backup.CMD - backup a computer to files, using ROBOCOPY and NTBACKUP
::
:: SYNOPSIS
::
::	ROBOCOPY_backup <target> <job_dir> <log_dir> <reporting_email> <drives...>
::
:: ARGUMENTS
::
::  target = backup files will be copyied/written to directories under this directory
::  job_dir = directory containing ROBOCOPY job (RCJ) files
::  log_dir = directory log file will be written to
::  reporting_email = email address to mail copy of log to when done, or "NONE" (without quotes) for no email
::  drives = one or more drive letters, *WITHOUT* the colons
:: 
:: PREREQUISITES
::
::  Target directory structure (subdirs SysState, C, D, etc.)  must exist
::
:: Various RCJ files
::  generic_all - global options which apply to all computers
::  generic_X - generic excludes which apply to drive letter X on all computers (one RCJ per drive letter)
::  %COMPUTERNAME%_X - optional; computer-specific excludes/options for drive letter X (one RCJ per computer per drive letter)

:: Utilities in PATH:
::  TEE (output copy)
::  GNU FIND utility (or compatible) as "GFIND"
::  RM (remove file)
::
:: If sending mail:
::  BLAT.EXE (mail tool) in PATH
::  %SMTPHOST% environment variable with name of SMTP mail relay server
::
:: NOTICES
:: 
:: Written by Ben Scott
:: Provided free-of-charge, "AS IS", WITHOUT WARRANTY.  USE AT YOUR OWN RISK.

SETLOCAL ENABLEEXTENSIONS
SET SELF=%0

:: ----------------------------------------------------------------------
:: build job ID

:: put current date and time into a variable
:: this serves as sort of a "job ID" for this run
:: format YYYY-MM-DD-HH-MM-SS

FOR /F "tokens=2-4 delims=/ " %%a IN ("%DATE%") DO SET MYDATE=%%c-%%a-%%b
FOR /F "tokens=1-3 delims=.: " %%a IN ("%TIME%") DO SET MYTIME=%%a-%%b-%%c
SET JOBID=%MYDATE%-%MYTIME%

:: ----------------------------------------------------------------------
:: command arguments

SET TARGET=%1
SHIFT
SET JOBDIR=%1
SHIFT
SET LOGDIR=%1
SHIFT
SET MAILTO=%1
SHIFT

:: do **NOT** call SHIFT for this or you will break the ROBOCOPY subroutines
SET FIRST_DRIVE=%1

IF NOT DEFINED TARGET (
	ECHO Missing argument
	ECHO Specify target directory [to write backups to] as next argument
	GOTO early_error
	)

IF NOT EXIST "%TARGET%" (
	ECHO Specified target directory does not exist: %TARGET%
	GOTO early_error
	)

IF NOT DEFINED JOBDIR (
	ECHO Missing argument
	ECHO Specify job directory as next argument
	GOTO early_error
	)

IF NOT EXIST "%JOBDIR%" (
	ECHO Specified job directory does not exist: %JOBDIR%
	GOTO early_error
	)

IF NOT DEFINED LOGDIR (
	ECHO Missing argument
	ECHO Specify log directory as next argument
	GOTO early_error
	)

IF NOT EXIST "%LOGDIR%" (
	ECHO Specified log directory does not exist: %LOGDIR%
	GOTO early_error
	)

IF NOT DEFINED MAILTO (
	ECHO Missing argument
	ECHO Specify email address to receive reports, or NONE
	GOTO early_error
	)

:: if MAILTO was set to NONE, clear it
:: after this, if MAILTO is defined, send mail; if not defined, do not send mail
IF MAILTO==NONE SET MAILTO=
IF MAILTO==none SET MAILTO=

IF DEFINED MAILTO IF NOT DEFINED SMTPHOST (
	ECHO Email report requested but %%SMTPHOST%% not defined
	GOTO early_error
	)

IF NOT DEFINED FIRST_DRIVE (
	ECHO Missing argument
	ECHO Specify drive letters to backup as remaining arguments
	ECHO Do not include colon : in drive letter arguments!
	GOTO early_error
	)

:: ----------------------------------------------------------------------
:: derived configuration

:: this will be *our* log file, not the one NTBACKUP creates, or that ROBOCOPY uses
SET LOGFILE=%LOGDIR%\%JOBID%.LOG

:: target file of NTBACKUP SysState backup-to-file
SET BKF=%TARGET%\SysState\SysState_%JOBID%.BKF

:: backup description
:: FIXME we tell NTBACKUP to use this, but it appears to ignore us
SET BKUP_DESC=%COMPUTERNAME%_%JOBID%

:: directory NTBACKUP keeps its log files in
SET NTLOGDIR=%USERPROFILE%\Local Settings\Application Data\Microsoft\Windows NT\NTBackup\data\

:: temporary robocopy log file (so we can filter it)
SET ROBOLOGFILE=%TEMP%\ROBOCOPY_%JOBID%.LOG

:: ----------------------------------------------------------------------
:: begin log file

CALL :logmsg Script started
CALL :logmsg SELF=%SELF%
CALL :logmsg CD=%CD%
CALL :logmsg COMPUTERNAME=%COMPUTERNAME%
CALL :logmsg JOBID=%JOBID%
CALL :logmsg TARGET=%TARGET%
CALL :logmsg JOBDIR=%JOBDIR%
CALL :logmsg LOGFILE=%LOGFILE%
CALL :logmsg BKF=%BKF%

:: ----------------------------------------------------------------------
:: make sure required tools are available

:: TEE copies stdout from a program to its stdout *and* to a file (Unix utility)
CALL :require_program TEE.EXE
IF ERRORLEVEL 1 GOTO finish

:: GNU find (or compatible; Unix utility), named GFIND.EXE to distinquish from Windows FIND.EXE
CALL :require_program GFIND.EXE
IF ERRORLEVEL 1 GOTO finish

:: RM removes files, called by GFIND (Unix utility)
CALL :require_program RM.EXE
IF ERRORLEVEL 1 GOTO finish

:: if mail is requested, we need BLAT to send it
IF DEFINED MAILTO (
	CALL :require_program BLAT.EXE
	IF ERRORLEVEL 1 GOTO finish
	)

:: ----------------------------------------------------------------------
:: sanity checks

IF EXIST "%ROBOLOGFILE%" (
	CALL :logmsg ROBOCOPY log file already exists!?!
	CALL :logmsg ROBOLOGFILE=%ROBOLOGFILE%
	CALL :logmsg Unexpected error!  Cannot continue!
	CALL :logmsg %SELF% aborting!
	GOTO finish
	)

:: ----------------------------------------------------------------------
:: call the backup routines

:: backup System State (especially registry and Active Directory)
CALL :run_NTBACKUP
CALL :process_NTBACKUP_log
CALL :purge_old_BKFs

:: backup plain files
:: run ROBOCOPY once for each 
:next_drive
SET NEXT_DRIVE=%1
SHIFT
IF DEFINED NEXT_DRIVE (
	CALL :run_ROBOCOPY %NEXT_DRIVE%
	GOTO next_drive
	)
CALL :process_ROBOCOPY_log

:: ----------------------------------------------------------------------
:: post-run operations and cleanup

:: if things start get going but then fall apart, GOTO here rather than exiting script
:finish

IF DEFINED MAILTO CALL :email_log_file

CALL :logmsg SELF=%SELF%
CALL :logmsg Script ending

:: exit script
EXIT /B

:: ======================================================================
:: subroutines

:: ----------------------------------------------------------------------
:run_NTBACKUP

CALL :logmsg Starting NTBACKUP...

NTBACKUP BACKUP SystemState ^
	/V:no /R:no /L:S /M normal ^
	/F "%BKF%" ^
	/J "%BKUP_DESC%" ^
	/N "BKF for %BKUP_DESC%" ^
	/D "%BKUP_DESC%"

CALL :logmsg NTBACKUP exited, ERRORLEVEL=%ERRORLEVEL%

:: return to CALLer
EXIT /B

:: ----------------------------------------------------------------------
:process_NTBACKUP_log

CALL :logmsg Processing NTBACKUP log file...

FOR /F "Tokens=*" %%a IN ('DIR "%NTLOGDIR%\backup*.log" /b /o:d') DO (SET NTLOGFILE=%%a)
CALL :logmsg NTLOGFILE=%NTLOGFILE%
CALL :logmsg Appending NTBACKUP log file to our log file...
ECHO ============================================================ >> "%LOGFILE%"
TYPE "%NTLOGDIR%\%NTLOGFILE%" >> "%LOGFILE%"
ECHO ============================================================ >> "%LOGFILE%"
	
:: return to CALLer
EXIT /B

:: ----------------------------------------------------------------------
:purge_old_BKFs

:: delete SysteState BKFs older than today
CALL :logmsg Purging old backup files:
gfind %TARGET%\SysState -daystart -type f -a -mtime +0 -a -iname SysState_*.BKF -a -print -a -exec rm {} ; | tee -a "%LOGFILE%"
CALL :logmsg Purge finished

:: return to CALLer
EXIT /B

:: ----------------------------------------------------------------------
:run_ROBOCOPY

:: reset
SET ROBO_OPTS=

:: logging
SET ROBO_OPTS=%ROBO_OPTS% /LOG+:%ROBOLOGFILE%

:: generic options which apply to all server backups, all drives
SET RCJ_GEN_ALL=%JOBDIR%\generic_all.RCJ
IF NOT EXIST %RCJ_GEN_ALL% (
	CALL :logmsg Missing RCJ: %RCJ_GEN_ALL%
	CALL :logmsg Aborting this ROBOCOPY run!
	:: return to CALLer
	EXIT /B 1
	)
SET ROBO_OPTS=%ROBO_OPTS% /JOB:%RCJ_GEN_ALL%

:: generic exclusions which apply to this drive on all computers
SET RCJ_GEN_DRV=%JOBDIR%\generic_%1.RCJ
IF NOT EXIST %RCJ_GEN_DRV% (
	CALL :logmsg Missing RCJ: %RCJ_GEN_DRV%
	CALL :logmsg Aborting this ROBOCOPY run!
	:: return to CALLer
	EXIT /B 1
	)
SET ROBO_OPTS=%ROBO_OPTS% /JOB:%RCJ_GEN_DRV%

:: exclusions specific to this drive on this server
SET RCJ_THS_DRV=%JOBDIR%\%COMPUTERNAME%_%1.RCJ
IF EXIST %RCJ_THS_DRV% SET ROBO_OPTS=%ROBO_OPTS% /JOB:%RCJ_THS_DRV%

CALL :logmsg Running ROBOCOPY for partition %1:
ROBOCOPY ^
	%ROBO_OPTS% ^
	%1:\ ^
	%TARGET%\%1

:: return to CALLer
EXIT /B

:: ----------------------------------------------------------------------
:process_ROBOCOPY_log

CALL :logmsg Processing ROBOCOPY log file...
ECHO ============================================================ >> "%LOGFILE%"
FINDSTR /V /C:"*EXTRA File" /C:"*EXTRA Dir" "%ROBOLOGFILE%" >> "%LOGFILE%"
ECHO ============================================================ >> "%LOGFILE%"

:: delete ROBOCOPY temp log file -- we are done with it now
DEL "%ROBOLOGFILE%"

IF EXIST "%ROBOLOGFILE%" CALL :logmsg WARNING: ROBOCOPY log file still exists: %ROBOLOGFILE%

:: return to CALLer
EXIT /B

:: ----------------------------------------------------------------------
:email_log_file

CALL :logmsg Emailing log file...
BLAT "%LOGFILE%" ^
	-server %SMTPHOST% ^
	-s "%COMPUTERNAME% ROBOCOPY backup report %JOBID%" ^
	-f "%USERNAME%@%COMPUTERNAME%" ^
	-to "%MAILTO%"
:: return to CALLer
EXIT /B

:: ----------------------------------------------------------------------
:early_error
:: called for early errors/aborts, before logging is working
ECHO Error during initialization!  Cannot continue!
ECHO %SELF% aborting
:: exit script
EXIT /B

:: ----------------------------------------------------------------------
:logmsg
SET LOGMSGTIME=[%DATE% %TIME%]
ECHO %LOGMSGTIME% %*
ECHO %LOGMSGTIME% %* >> "%LOGFILE%"
:: return to CALLer
EXIT /B

:: ----------------------------------------------------------------------
:require_program
:: makes sure a program file exists in PATH
:: returns success if true, failure if false
:: we cannot do the actual script abort here (would get interpreted as subroutine CALL return)
SET require_program_test_path=
FOR %%i IN (%1) DO SET require_program_test_path=%%~$PATH:i
IF NOT DEFINED require_program_test_path (
	CALL :logmsg Missing required tool: %1
	EXIT /B 1
	)
EXIT /B 0

:: ----------------------------------------------------------------------
:: END OF FILE
:: ----------------------------------------------------------------------
